import torch
from torch import nn
import numpy as np

gpu_num=1

class IMVTensorLSTM(torch.jit.ScriptModule):
    __constants__ = ["n_units", "input_dim"]

    def __init__(self, input_dim, output_dim, n_units, init_std=0.02, gpu_num=gpu_num):
        super().__init__()
        self.U_j = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)
        self.U_i = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)
        self.U_f = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)
        self.U_o = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)
        self.W_j = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std)
        self.W_i = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std)
        self.W_f = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std)
        self.W_o = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std)
        self.b_j = nn.Parameter(torch.randn(input_dim, n_units) * init_std)
        self.b_i = nn.Parameter(torch.randn(input_dim, n_units) * init_std)
        self.b_f = nn.Parameter(torch.randn(input_dim, n_units) * init_std)
        self.b_o = nn.Parameter(torch.randn(input_dim, n_units) * init_std)
        self.F_alpha_n = nn.Parameter(torch.randn(input_dim, n_units, 1) * init_std)
        self.F_alpha_n_b = nn.Parameter(torch.randn(input_dim, 1) * init_std)
        self.F_beta = nn.Linear(2 * n_units, 1)
        self.Phi = nn.Linear(2 * n_units, output_dim)
        self.n_units = n_units
        self.input_dim = input_dim
        self.gpu_num = gpu_num
        self.device = torch.device("cuda:%d" % self.gpu_num if torch.cuda.is_available() else "cpu")
    @torch.jit.script_method
    def forward(self, x):
        h_tilda_t = torch.zeros(x.shape[0], self.input_dim, self.n_units).to(self.device)
        c_tilda_t = torch.zeros(x.shape[0], self.input_dim, self.n_units).to(self.device)
        outputs = torch.jit.annotate(List[Tensor], [])
        for t in range(x.shape[1]):
            # eq 1
            j_tilda_t = torch.tanh(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_j) + \
                                   torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_j) + self.b_j)
            # eq 5
            i_tilda_t = torch.sigmoid(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_i) + \
                                      torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_i) + self.b_i)
            f_tilda_t = torch.sigmoid(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_f) + \
                                      torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_f) + self.b_f)
            o_tilda_t = torch.sigmoid(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_o) + \
                                      torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_o) + self.b_o)
            # eq 6
            c_tilda_t = c_tilda_t * f_tilda_t + i_tilda_t * j_tilda_t
            # eq 7
            h_tilda_t = (o_tilda_t * torch.tanh(c_tilda_t))
            outputs += [h_tilda_t]
        outputs = torch.stack(outputs)
        outputs = outputs.permute(1, 0, 2, 3)
        # eq 8
        alphas = torch.tanh(torch.einsum("btij,ijk->btik", outputs, self.F_alpha_n) + self.F_alpha_n_b)
        alphas = torch.exp(alphas)
        alphas = alphas / torch.sum(alphas, dim=1, keepdim=True)
        g_n = torch.sum(alphas * outputs, dim=1)
        hg = torch.cat([g_n, h_tilda_t], dim=2)
        mu = self.Phi(hg)
        betas = torch.tanh(self.F_beta(hg))
        betas = torch.exp(betas)
        betas = betas / torch.sum(betas, dim=1, keepdim=True)
        mean = torch.sum(betas * mu, dim=1)

        return mean, alphas, betas


class IMVFullLSTM(torch.jit.ScriptModule):
    __constants__ = ["n_units", "input_dim"]

    def __init__(self, input_dim, output_dim, n_units, init_std=0.02, gpu_num=gpu_num):
        super().__init__()
        # x: input_dim, y:output_dim, h:n_units
        # torch.randn initializes tensors filled with random values, then init_std scales them by 0.02
        self.U_j = nn.Parameter(torch.randn(input_dim, 1, n_units) * init_std)  #[x, 1, h]
        self.W_j = nn.Parameter(torch.randn(input_dim, n_units, n_units) * init_std) #[x,h,h]
        self.b_j = nn.Parameter(torch.randn(input_dim, n_units) * init_std) #[x,h]
        self.W_i = nn.Linear(input_dim * (n_units + 1), input_dim * n_units) #maps x.(h+1)-->x.h
        self.W_f = nn.Linear(input_dim * (n_units + 1), input_dim * n_units) #maps x.(h+1)-->x.h
        self.W_o = nn.Linear(input_dim * (n_units + 1), input_dim * n_units) #maps x.(h+1)-->x.h
        self.F_alpha_n = nn.Parameter(torch.randn(input_dim, n_units, 1) * init_std) #[x,h,1]
        self.F_alpha_n_b = nn.Parameter(torch.randn(input_dim, 1) * init_std) #[x,1]
        self.F_beta = nn.Linear(2 * n_units, 1) #maps 2h-->1
        self.Phi = nn.Linear(2 * n_units, output_dim) #maps 2h-->y
        self.n_units = n_units
        self.input_dim = input_dim
        self.gpu_num = gpu_num
        self.device = torch.device("cuda:%d" % self.gpu_num if torch.cuda.is_available() else "cpu")

    @torch.jit.script_method
    def forward(self, x):
        # x shape: [b,d,x]
        # d: depth (represent timesteps), b:batch size
        # initializing tensors filled with zeros
        h_tilda_t = torch.zeros(x.shape[0], self.input_dim, self.n_units, device=self.device) #[b,x,h]
        c_t = torch.zeros(x.shape[0], self.input_dim * self.n_units, device=self.device) #[b,x.h]
        outputs = torch.jit.annotate(List[Tensor], []) #outputs will be a list of tensors, this just specifies the type contained within the list
        for t in range(x.shape[1]): #for the elements along the depth, so timesteps
            # eq 1
            #this is the hidden state update
            # update from hidden state from previous timesteps
            j_tilda_t = torch.tanh(torch.einsum("bij,ijk->bik", h_tilda_t, self.W_j) + \
                                   torch.einsum("bij,jik->bjk", x[:, t, :].unsqueeze(1), self.U_j) + self.b_j).to(self.device) #update from the current input and then bias term
            # the same bias is applied to each member in the batch
            # j_tilda_t shape [b,x,h]
            # inp contains both hidden state and current input information
            # concatenate x at the current timestep with h_tilda_t along dim=1 the h dimension 
            inp = torch.cat([x[:, t, :], h_tilda_t.reshape(h_tilda_t.shape[0], -1)], dim=1).to(self.device) #[b, x.(h+1)
            # eq 2
            # W_i W_f W_o are reshaping, but also applying individual weight matrices
            i_t = torch.sigmoid(self.W_i(inp)).to(self.device) #[b, x.h]
            f_t = torch.sigmoid(self.W_f(inp)).to(self.device) #[b, x.h]
            o_t = torch.sigmoid(self.W_o(inp)).to(self.device) #[b, x.h]
            # eq 3
            # the memory cell state
            c_t = c_t * f_t + i_t * j_tilda_t.reshape(j_tilda_t.shape[0], -1).to(self.device) #[b,x.h]
            # eq 4
            # now the hidden state is updated
            h_tilda_t = (o_t * torch.tanh(c_t)).reshape(h_tilda_t.shape[0], self.input_dim, self.n_units).to(self.device) #[b,x,h]
            # the hidden state also forms the "output"
            # outputs is a sequence of hidden states
            outputs += [h_tilda_t]
        outputs = torch.stack(outputs).to(self.device) #[d,b,x,h]
        outputs = outputs.permute(1, 0, 2, 3).to(self.device) #[b,d,x,h]
        # eq 8
        # temporal variable attention
        alphas = torch.tanh(torch.einsum("btij,ijk->btik", outputs, self.F_alpha_n) + self.F_alpha_n_b).to(self.device) #[b,d,x,1]
        # applying a softmax (normalized exponential) function
        # this normalizes along the the depth, so among timesteps for each variable seperately
        alphas = torch.exp(alphas).to(self.device)
        alphas = alphas / torch.sum(alphas, dim=1, keepdim=True).to(self.device) #[b,d,x,1]
        # applying the temporal attention and then summing along the depth
        g_n = torch.sum(alphas * outputs, dim=1).to(self.device) #[b,x,h]
        # concatenating the temporal attention weighted hidden state and the hidden state itself only from the last timestep
        hg = torch.cat([g_n, h_tilda_t], dim=2).to(self.device) #[b,x,2h]
        mu = self.Phi(hg).to(self.device) #[b,x,y]
        # now variable attention (not temporal)
        betas = torch.tanh(self.F_beta(hg)).to(self.device) #[b,x,1]
        # softmax again, but now normalizing among variables
        betas = torch.exp(betas).to(self.device)
        betas = betas / torch.sum(betas, dim=1, keepdim=True).to(self.device) #[b,x,1]
        # now calculating the final output 
        # this considers both the temporal variable attention and non-temporal attention
        # sums along the variables
        mean = torch.sum(betas * mu, dim=1).to(self.device) #[b,1]
        return mean, alphas, betas
